# HAOT modules
from .aerodynamics import *
from .quantum_mechanics import *
from .optics import *
from .constants import *
from .conversions import *

# Printing Version
from importlib.metadata import version

# Print Version
__version__ = version("haot")



"""
    Date:   11/26/2024
    Author: Martin E. Liza
    File:   quantum_mechanics.py
    Def:    Contains Quantum Mechanics functions.
"""

import molmass
import numpy as np
import scipy.constants as s_consts
from haot import constants as constants_tables
from haot import conversions


def zero_point_energy(molecule: str) -> float:
    """
    Calculates zero-point energy (ZPE) using spectroscopy constants for
    diatomic molecules

    Parameters:
        molecule: NO+, N2+, O2+, NO, N2, O2, H2

    Returns:
        zero point energy in [cm^-1]

    Examples:
        >> zero_point_energy('O2')

    Reference:
        Experimental Vibrational Zero-Point Energies Diatomic Molecules
        (https://doi.org/10.1063/1.2436891)
    """
    # Unit Test
    if molecule not in ["NO+", "N2+", "O2+", "NO", "N2", "O2", "H2"]:
        raise ValueError("This function only supports NO+, N2+, O2+, NO, N2, O2, H2")
    spectroscopy_const = constants_tables.spectroscopy_constants(molecule)

    scope_var = spectroscopy_const["alpha_e"]
    scope_var *= spectroscopy_const["omega_e"]
    scope_var /= spectroscopy_const["B_e"]

    zpe = spectroscopy_const["omega_e"] / 2
    zpe -= spectroscopy_const["omega_xe"] / 2
    zpe += spectroscopy_const["omega_ye"] / 8
    zpe += spectroscopy_const["B_e"] / 4
    zpe += scope_var / 12
    zpe += scope_var**2 / (144 * spectroscopy_const["B_e"])
    return zpe  # [1/cm]


def vibrational_partition_function(
    vibrational_number: int, temperature_K: float, molecule: str
) -> float:
    """
    Calculates the vibrational partition function base in the harmonic
    terms only for diatomic molecules

    Parameters:
        vibrational_number: vibrational quantum number (has to be positive)
        temperature_K: reference temperature in [K]
        molecule: NO+, N2+, O2+, NO, N2, O2

    Returns:
        vibrational partition function

    Examples:
        >> vibrational_partition_function(2, 350.0, 'N2')
    """
    # Unit Test
    if not isinstance(vibrational_number, int):
        raise ValueError("Vibrational quantum number should be a positive integer")
    if vibrational_number < 0.0:
        raise ValueError("Vibrational number should be a positive integer!")
    if temperature_K < 0.0:
        raise ValueError("Temperature should be positive!")
    if molecule not in ["NO+", "N2+", "O2+", "NO", "N2", "O2", "H2"]:
        raise ValueError("This function only supports NO+, N2+, O2+, NO, N2, O2, H2")
    z_vib = 0.0
    for v in range(vibrational_number + 1):
        z_vib += boltzmann_factor(
            temperature_K,
            molecule,
            vibrational_number=v,
            rotational_number=None,
            born_opp_flag=False,
        )
    return z_vib


def rotational_partition_function(
    rotational_number: int, temperature_K: float, molecule: str
) -> float:
    """
    Calculates the rotational partition function base in the harmonic
    terms only for diatomic molecules

    Parameters:
        rotational_number: rotational quantum number (has to be positive)
        temperature_K: reference temperature in [K]
        molecule: NO+, N2+, O2+, NO, N2, O2

    Returns:
        rotational partition function

    Examples:
        >> rotational_partition_function(2, 350.0, 'N2')
    """
    # Unit Test
    if not isinstance(rotational_number, int):
        raise ValueError("Vibrational quantum number should be a positive integer")
    if rotational_number < 0.0:
        raise ValueError("Rotational number should be a positive integer!")
    if temperature_K < 0.0:
        raise ValueError("Temperature should be positive!")
    if molecule not in ["NO+", "N2+", "O2+", "NO", "N2", "O2", "H2"]:
        raise ValueError("This function only supports NO+, N2+, O2+, NO, N2, O2, H2")
    z_rot = 0.0
    for j in range(rotational_number + 1):
        z_rot += boltzmann_factor(
            temperature_K,
            molecule,
            vibrational_number=None,
            rotational_number=j,
            born_opp_flag=False,
        )
    return z_rot


def born_oppenheimer_partition_function(
    vibrational_number: int, rotational_number: int, temperature_K: float, molecule: str
) -> float:
    """
    Calculates the partition function using the Born-Oppenheimer approximation

    Parameters:
        vibrational_number: vibrational quantum number (has to be positive)
        rotational_number: rotational quantum number (has to be positive)
        temperature_K: reference temperature in [K]
        molecule: NO+, N2+, O2+, NO, N2, O2

    Returns:
        partition function using the Born-Oppenheimer approximations

    Examples:
        >> born_oppenheimer_partition_function(2, 4, 500.0, 'O2')
    """
    # Unit Test
    if not isinstance(vibrational_number, int):
        raise ValueError("Vibrational quantum number should be a positive integer")
    if vibrational_number < 0.0:
        raise ValueError("Vibrational number should be a positive integer!")
    if not isinstance(rotational_number, int):
        raise ValueError("Vibrational quantum number should be a positive integer")
    if rotational_number < 0.0:
        raise ValueError("Rotational number should be a positive integer!")
    if temperature_K < 0.0:
        raise ValueError("Temperature should be positive!")
    if molecule not in ["NO+", "N2+", "O2+", "NO", "N2", "O2", "H2"]:
        raise ValueError("This function only supports NO+, N2+, O2+, NO, N2, O2, H2")
    z_bo = 0.0
    for j in range(rotational_number + 1):
        for v in range(vibrational_number + 1):
            z_bo += boltzmann_factor(
                temperature_K,
                molecule,
                vibrational_number=v,
                rotational_number=j,
                born_opp_flag=True,
            )
    return z_bo


def potential_dunham_coef_012(molecule: str) -> tuple[float, float, float]:
    """
    Calculates the 0th, 1st, and 2nd Dunham potential coefficients for diatomic
    molecules

    Parameters:
        molecule: NO+, N2+, O2+, NO, N2, O2

    Returns:
       Dunham potential coefficients 0, 1, and 2

    Examples:
        >> [a_0, a_1, a_2] = potential_dunham_coef_012('N2')

    Reference:
        Dunham potential energy coefficients of the hydrogen halides and carbon
        monoxide (https://doi.org/10.1016/0022-2852(76)90323-4)

        Anharmonic Potential Constants and Their Dependence upon Bond Length
        (https://doi.org/10.1063/1.1731952)
    """
    spectroscopy_const = constants_tables.spectroscopy_constants(molecule)
    a_0 = spectroscopy_const["omega_e"] ** 2 / (4 * spectroscopy_const["B_e"])
    a_1 = -(
        spectroscopy_const["alpha_e"]
        * spectroscopy_const["omega_e"]
        / (6 * spectroscopy_const["B_e"] ** 2)
        + 1
    )
    a_2 = (5 / 4) * a_1**2 - (2 / 3) * (
        spectroscopy_const["omega_xe"] / spectroscopy_const["B_e"]
    )
    return (a_0, a_1, a_2)


def potential_dunham_coeff_m(a_1: float, a_2: float, m: int) -> float:
    """
    Calculates higher order Dunham potential coefficients

    Parameters:
        a_1: Dunham potential coefficient
        a_2: Dunham potential coefficient
        m:   desired Dunham potential coefficient

    Returns:
       Dunham potential coefficient at m

    Examples:
        >> a_3 = potential_dunham_coef_m(a_1, a_2, 3)

    Reference:
        A recursion formula for the coefficients of Dunham potential (https://doi.org/10.1016/j.theochem.2003.12.003)
    """

    tmp = (12 / a_1) ** (m - 2)
    tmp *= 2 ** (m + 1) - 1
    tmp *= (a_2 / 7) ** (m - 1)
    for i in range(m - 2):
        tmp *= 1 / (m + 2 - i)

    return tmp


def boltzmann_factor(
    temperature_K: float,
    molecule: str,
    vibrational_number: int = None,
    rotational_number: int = None,
    born_opp_flag: bool = False,
) -> float:
    """
    Calculates the Boltzmann factor at a given vibrational quantum number,
    and/or rotational quantum number. If the born_opp_flag is provided, it will
    calculate the Boltzmann factor using the Born-Oppenheimer approximation

    Parameters:
        temperature_K: reference temperature in [K]
        molecule: NO+, N2+, O2+, NO, N2, O2
        vibrational_number: vibrational quantum number (has to be positive)
        rotational_number: rotational quantum number (has to be positive)
        born_opp_flag: uses the Born-Oppenheimer approximation, False (default)

    Returns:
        Boltzmann factor at given temperature, rotational and/or vibrational quantum number

    Examples:
        >> boltzmann_factor(500.0, 'O2', 3)

        >> boltzmann_factor(500.0, 'O2', 3, 2)

        >> boltzmann_factor(500.0, 'O2', 3, 1, True)
    """
    # Initialize energy terms, degeneracy and thermal beta
    energy_vib_k = 0
    energy_rot_k = 0
    degeneracy_rotation = 1
    thermal_beta = 1 / (s_consts.k * temperature_K)

    # Calculates Energy levels
    if not born_opp_flag:
        if vibrational_number is not None:
            energy_vib_k = vibrational_energy_k(vibrational_number, molecule)
        if rotational_number is not None:
            energy_rot_k = rotational_energy_k(rotational_number, molecule)
            degeneracy_rotation = 2 * rotational_number + 1
        tot_energy = conversions.wavenumber_to_joules(energy_vib_k + energy_rot_k)
    else:
        degeneracy_rotation = 2 * rotational_number + 1
        tot_energy = conversions.wavenumber_to_joules(
            born_oppenheimer_approximation(
                vibrational_number, rotational_number, molecule
            )
        )
    return degeneracy_rotation * np.exp(-tot_energy * thermal_beta)


def boltzmann_distribution(
    temperature_K: float,
    molecule: str,
    vibrational_number: int = None,
    rotational_number: int = None,
    born_opp_flag: bool = False,
) -> np.ndarray:
    """
    Calculates the Boltzmann distribution function at a given vibrational quantum number,
    and/or rotational quantum number. If the born_opp_flag is provided, it will
    calculate the Boltzmann factor using the Born-Oppenheimer approximation

    Parameters:
        temperature_K: reference temperature in [K]
        molecule: NO+, N2+, O2+, NO, N2, O2
        vibrational_number: vibrational quantum number (has to be positive)
        rotational_number: rotational quantum number (has to be positive)
        born_opp_flag: uses the Born-Oppenheimer approximation, False (default)

    Returns:
        Boltzmann distribution [vibrational_number, rotational_number]

    Examples:
        >> boltzmann_distribution(500.0, 'O2', 3, 2, true)
    """
    # Calculates partition functions if vibrational or rotational numbers are provided
    if not born_opp_flag:
        z_rot = 1
        z_vib = 1
        if vibrational_number is not None:
            z_vib = vibrational_partition_function(
                vibrational_number, temperature_K, molecule
            )
        if rotational_number is not None:
            z_rot = rotational_partition_function(
                rotational_number, temperature_K, molecule
            )
        z_tot = z_rot * z_vib
    else:
        z_tot = born_oppenheimer_partition_function(
            vibrational_number, rotational_number, temperature_K, molecule
        )

    # Create the distribution array based on the inputs provided
    if vibrational_number and rotational_number:
        tmp = np.zeros([vibrational_number + 1, rotational_number + 1])
        for j in range(rotational_number + 1):
            for v in range(vibrational_number + 1):
                tmp[v][j] = boltzmann_factor(
                    temperature_K=temperature_K,
                    molecule=molecule,
                    vibrational_number=v,
                    rotational_number=j,
                    born_opp_flag=born_opp_flag,
                )
    elif vibrational_number:
        tmp = np.zeros(vibrational_number + 1)
        for v in range(vibrational_number + 1):
            tmp[v] = boltzmann_factor(
                temperature_K=temperature_K,
                molecule=molecule,
                vibrational_number=v,
                rotational_number=None,
                born_opp_flag=False,
            )
    elif rotational_number:
        tmp = np.zeros(rotational_number + 1)
        for j in range(rotational_number + 1):
            tmp[j] = boltzmann_factor(
                temperature_K=temperature_K,
                molecule=molecule,
                vibrational_number=None,
                rotational_number=j,
                born_opp_flag=False,
            )
    return tmp / z_tot


def born_oppenheimer_approximation(
    vibrational_number: int, rotational_number: int, molecule: str
) -> float:
    """
    Calculates the energy at a rotational and vibrational quantum number,
    using the Born-Oppenheimer approximation

    Parameters:
        molecule: NO+, N2+, O2+, NO, N2, O2
        vibrational_number: vibrational quantum number (has to be positive),
        rotational_number: rotational quantum number (has to be positive)

    Returns:
        Vibro-rotational energy at given vibrational and rotational quantum
        numbers in [cm^-1]

    Examples:
        >> born_oppenheimer_approximation(2,3,'N2')
    """
    spectroscopy_constants = constants_tables.spectroscopy_constants(molecule)

    vib_levels = vibrational_number + 1 / 2
    rot_levels = rotational_number * (rotational_number + 1)

    # Harmonic vibration and rotation terms
    harmonic = spectroscopy_constants["omega_e"] * vib_levels
    harmonic += spectroscopy_constants["B_e"] * rot_levels

    # Anharmonic vibration and rotation terms
    anharmonic = spectroscopy_constants["omega_xe"] * vib_levels**2
    anharmonic += spectroscopy_constants["D_e"] * rot_levels**2

    # Interaction between vibration and rotation modes
    interaction = spectroscopy_constants["alpha_e"] * vib_levels * rot_levels

    return harmonic - anharmonic - interaction  # [cm^1]


def vibrational_energy_k(vibrational_number: int, molecule: str) -> float:
    """
    Calculates the vibrational energy at a given vibrational quantum number,
    using for the harmonic terms

    Parameters:
        vibrational_number: vibrational quantum number (has to be positive)
        molecule: NO+, N2+, O2+, NO, N2, O2

    Returns:
        harmonic terms of the vibrational energy in [cm^-1]

    Examples:
        >> vibrational_energy_k(2, 'N2')
    """
    spectroscopy_constants = constants_tables.spectroscopy_constants(molecule)
    # Calculates the vibrational energy in units of wave number
    vib_levels = vibrational_number + 1 / 2
    return spectroscopy_constants["omega_e"] * vib_levels  # [cm^-1]


def rotational_energy_k(rotational_number: int, molecule: str) -> float:
    """
    Calculates the rotational energy at a given rotational quantum number,
    using for the harmonic terms

    Parameters:
        rotational_number: rotational quantum number (has to be positive)
        molecule: NO+, N2+, O2+, NO, N2, O2

    Returns:
        harmonic terms of the rotational energy in [cm^-1]

    Examples:
        >> rotational_energy_k(2, 'N2')
    """
    spectroscopy_constants = constants_tables.spectroscopy_constants(molecule)
    # Calculates the rotational energy in units of wave number
    rot_levels = rotational_number * (rotational_number + 1)
    return spectroscopy_constants["B_e"] * rot_levels  # [cm^-1]


def reduced_mass_kg(molecule_1: str, molecule_2: str) -> float:
    """
    Calculates the reduced mass in [kg]

    Parameters:
        molecule_1: name of molecule one
        molecule_2: name of molecule two

    Returns:
        reduced mass in [kg]

    Examples:
        >> reduced_mass_kg('N', 'O')
    """
    m_1 = molmass.Formula(molecule_1).mass
    m_2 = molmass.Formula(molecule_2).mass
    mu = m_1 * m_2 / (m_1 + m_2)

    return conversions.molar_mass_to_kilogram(mu)


def molecular_spring_constant(molecule: str) -> float:
    """
    Calculates the molecular spring constant in [N/m]

    Parameters:
        molecule: NO+, N2+, O2+, NO, N2, O2

    Returns:
        spring force constant in [N/m]

    Examples:
        >> molecular_spring_constant('N2')
    """
    # Split masses
    m_1 = molecule[0]
    m_2 = m_1 if molecule[1] == str(2) else molecule[1]
    spectroscopy_const = constants_tables.spectroscopy_constants(molecule)
    mass_kg = reduced_mass_kg(m_1, m_2)
    freq_rps = conversions.wavenumber_to_angular_frequency(
        spectroscopy_const["omega_e"]
    )

    return mass_kg * freq_rps**2


# TODO: Missing Translational Energy
def tranlational_energy(principal_number_x, principal_number_y, principal_number_z):
    print("TODO: Missing implementation of this function")



"""
    Date:   12/11/2024
    Author: Martin E. Liza
    File:   conversions.py
    Def:    Contains conversion functions.
"""

import numpy as np
import molmass
import scipy.constants as s_consts


def polarizability_cgs_to_si(polarizability_cgs: float) -> float:
    """
    Converts volumetric polarizability (CGS) to atomic polarizability (SI)

    Parameters:
        polarizability_cgs: volumetric polarizability in [cm^3]

    Returns:
        atomic polarizability in [Fm^2]
    """
    return polarizability_cgs * 4 * np.pi * s_consts.epsilon_0 * 1e-6


def polarizability_si_to_cgs(polarizability_si: float) -> float:
    """
    Converts atomic polarizability (SI) to volumetric polarizability (CGS)

    Parameters:
        polarizability_si: atomic polarizability in [Fm^2]

    Returns:
        volumetric polarizability in [cm^3]
    """
    return polarizability_si * 1e6 / (4 * np.pi * s_consts.epsilon_0)


def wavenumber_to_electronvolt(wavenumber_cm: float) -> float:
    """
    Converts wavenumber to electron-volt

    Parameters:
        wavenumber_cm: energy in [cm^-1]

    Returns:
        energy in [eV]
    """
    return wavenumber_to_joules(wavenumber_cm) / s_consts.eV


def wavenumber_to_joules(wavenumber_cm: float) -> float:
    """
    Converts wavenumber to Joules

    Parameters:
        wavenumber_cm: energy in [cm^-1]

    Returns:
        energy in [J]
    """
    return wavenumber_cm * s_consts.c * 100 * s_consts.h


def wavenumber_to_angular_frequency(wavenumber_cm: float) -> float:
    """
    Converts wavenumber to [rads/seconds]

    Parameters:
        wavenumber_cm: energy in [cm^-1]

    Returns:
        frequency in [rads/seconds]
    """
    return 2 * np.pi * s_consts.c * wavenumber_cm * 100


def mass_density_to_molar_density(mass_density: float, molecule: str) -> float:
    """
    Converts mass_density [kg/m3] to molar density [particles/m3]

    Parameters:
        mass_density: density in [kg/m3]
        molecule: molecule of the mass density

    Returns:
        molar density in [particles/m3]
    """
    return mass_density * s_consts.N_A / molmass.Formula(molecule).mass * 1e3


def molar_mass_to_kilogram(molar_mass_gmol: float) -> float:
    """
    Converts molar mass [g/mol] to mass [kg]

    Parameters:
        molar_mass_gmol: mass in [g/mol]

    Returns:
        mass in [kg]
    """
    return molar_mass_gmol * 1e-3 / s_consts.N_A



"""
    Date:   03/26/2023
    Author: Martin E. Liza
    File:   optics.py
    Def:    Contains aero optics functions.
"""

from ambiance import Atmosphere
import molmass
import numpy as np
import scipy.constants as s_consts
from haot import constants as constants_tables
from haot import quantum_mechanics as quantum
from haot import conversions


def index_of_refraction_density_temperature(
    temperature_K: float,
    mass_density: float,
    molecule: str = "Air",
    wavelength_nm: float = 633.0,
) -> dict[str, float]:
    """
    Calculates dilute and dense index of refraction as a
    function of mass density and temperature.
    Uses Kerl approximation for polarizability

    Parameters:
        temperature_K: reference temperature in [K]
        mass density in [kg/m^3]
        molecule: H2, N2, O2, Air(default)
        wavelength_nm: signal's wavelength in [nm], 633(default) [nm]

    Returns:
        dict: A dictionary containing
            - dilute: dilute index of refraction
            - dense: dense index of refraction
    """
    # Checks
    if molecule not in ["Air", "H2", "N2", "O2"]:
        raise ValueError("This function only supports Air, H2, N2 or O2")
    if type(temperature_K) is float and temperature_K < 0:
        raise ValueError("Temperature must be greater than 0 Kelvin!")
    if type(temperature_K) is np.ndarray and (temperature_K < 0).any():
        raise ValueError("Temperature must be greater than 0 Kelvin!")
    if wavelength_nm <= 0:
        raise ValueError("Wavelength must be greater than 0 nanometers!")
    # Calculates polarizability using Kerl
    pol_kerl_air_m3 = kerl_polarizability_temperature(
        temperature_K, "Air", wavelength_nm
    )
    pol_kerl_SI = conversions.polarizability_cgs_to_si(pol_kerl_air_m3 * 1e6)

    if molecule == "Air":
        molar_mass_air = (
            0.78 * molmass.Formula("O2").mass
            + 0.21 * molmass.Formula("N2").mass
            + 0.01 * molmass.Formula("Ar").mass
        )
        molar_density = mass_density * s_consts.N_A / molar_mass_air * 1e3
    else:
        molar_density = conversions.mass_density_to_molar_density(
            mass_density, molecule
        )
    tot_pol_molar = molar_density * pol_kerl_SI
    n_return = {}
    n_return["dilute"] = 1 + tot_pol_molar / (2 * s_consts.epsilon_0)
    n_temp = tot_pol_molar / (3 * s_consts.epsilon_0)
    n_return["dense"] = ((2 * n_temp + 1) / (1 - n_temp)) ** 0.5

    return n_return


def index_of_refraction(mass_density_dict: dict[str, float]) -> dict[str, float]:
    """
    Calculates dilute and dense index of refraction as a
    function of mass density

    Parameters:
        mass density dictionary in [kg/m^3]
        keys should be elements alone. Ex [N2, O2, O, N, NO]


    Returns:
        dict: A dictionary containing
            - dilute: dilute index of refraction
            - dense: dense index of refraction
    """
    # Unit Test
    if not isinstance(mass_density_dict, dict):
        raise ValueError("Mass density should be a dictionary!")
    if mass_density_dict.keys() not in ["N2", "O2", "O", "N", "NO"]:
        raise ValueError("Keys should be named: N2, O2, O, N, NO")

    pol_consts = constants_tables.polarizability()  # [cm3]
    molar_density = {
        key: conversions.mass_density_to_molar_density(value, key)
        for key, value in mass_density_dict.items()
    }
    # a_i * N_i
    n_const = {
        key: conversions.polarizability_cgs_to_si(pol_consts[key]) * molar_density[key]
        for key in mass_density_dict.keys()
    }
    # Sum (a_i N_i)
    tot_pol_molar = sum(n_const.values())

    # Calculates dilute and dense index of refraction
    n_return = {}
    n_return["dilute"] = 1 + tot_pol_molar / (2 * s_consts.epsilon_0)
    n_temp = tot_pol_molar / (3 * s_consts.epsilon_0)
    n_return["dense"] = ((2 * n_temp + 1) / (1 - n_temp)) ** 0.5

    return n_return


def permittivity_material(index_of_refraction: float) -> float:
    """
    Calculates the permittivity of the material for a linear dielectric.

    Parameters:
        index_of_refraction: index of refraction

    Returns:
        material's permittivity in [F/m]

    Reference:
        Introduction to Electrodynamics, 4th (Griffiths D., DOI:
        10.1017/9781108333511)

    """
    # Unit Test
    if not (
        isinstance(index_of_refraction, np.ndarray)
        or isinstance(index_of_refraction, float)
        or isinstance(index_of_refraction, int)
    ):
        raise ValueError(
            "Index of refraction must be a numpy.ndarray, float or integer"
        )
    if type(index_of_refraction) is float and index_of_refraction < 0:
        raise ValueError("Index of refraction must be greater than 0!")
    if type(index_of_refraction) is np.ndarray and (index_of_refraction < 0).any():
        raise ValueError("Index of must be greater than 0!")

    # n ~ sqrt(e_r), Eq. 4.33
    return s_consts.epsilon_0 * index_of_refraction**2


def electric_susceptibility(index_of_refraction: float) -> float:
    """
    Calculates the electric susceptibility for a linear dielectric.

    Parameters:
        index_of_refraction: index of refraction

    Returns:
        electric susceptibility in [ ]

    Reference:
        Introduction to Electrodynamics, 4th (Griffiths D., DOI:
        10.1017/9781108333511)
    """
    # Unit Test
    if not (
        isinstance(index_of_refraction, np.ndarray)
        or isinstance(index_of_refraction, float)
        or isinstance(index_of_refraction, int)
    ):
        raise ValueError(
            "Index of refraction must be a numpy.ndarray, float or integer"
        )
    if type(index_of_refraction) is float and index_of_refraction < 0:
        raise ValueError("Index of refraction must be greater than 0!")
    if type(index_of_refraction) is np.ndarray and (index_of_refraction < 0).any():
        raise ValueError("Index of must be greater than 0!")
    # Eq 4.34
    return index_of_refraction**2 - 1


def optical_path_length(index_of_refraction: float, distance: float) -> float:
    """
    Calculates the optical path length

    Parameters:
        index_of_refraction: index of refraction
        distance: length

    Returns:
        Optical Path Length in units of distance
    """
    # Unit Test
    if not (
        isinstance(index_of_refraction, np.ndarray)
        or isinstance(index_of_refraction, float)
        or isinstance(index_of_refraction, int)
    ):
        raise ValueError(
            "Index of refraction must be a numpy.ndarray, float or integer"
        )
    if type(index_of_refraction) is float and index_of_refraction < 0:
        raise ValueError("Index of refraction must be greater than 0!")
    if type(index_of_refraction) is np.ndarray and (index_of_refraction < 0).any():
        raise ValueError("Index of must be greater than 0!")
    if type(distance) is float and distance < 0:
        raise ValueError("Distance must be greater than 0!")
    if type(distance) is np.ndarray and (distance < 0).any():
        raise ValueError("Distance of must be greater than 0!")
    if np.shape(index_of_refraction) != np.shape(distance):
        raise ValueError("Index of refraction and distance must have the same length")
    index_avg = 0.5 * (index_of_refraction[:-1] + index_of_refraction[1:])
    return np.mean(np.cumsum(index_avg * np.diff(distance)))


def optical_path_difference_rms(opd: float, avg_ax: int = 0) -> float:
    """
    Calculates the optical path difference RMS.

    Parameters:
        opd: Optical Path Difference
        avg_ax: axis where average is performed, 0 (default)

    Returns
        Optical Path Difference Root-Mean-Squared
    """
    # Validate the input array
    if not isinstance(opd, np.ndarray):
        raise ValueError("opd must be a numpy array")

    if avg_ax not in [0, 1, 2, 3]:
        raise ValueError("avg_ax must be one of [0, 1, 2, 3]")
    return np.sqrt(np.mean((opd - np.mean(opd, axis=avg_ax, keepdims=True)) ** 2))


def phase_variance(opd_rms: float, wavelength_nm: float) -> float:
    """
    Calculates phase variance.

    Parameters:
        opd_rm: Optical Path Difference RMS in units of [m]
        wavelength_nm: Wavelength of light in units of [nm]

    Returns:
        Phase difference, unit-less
    """
    return (2 * np.pi * opd_rms / (wavelength_nm * 1e-9)) ** 2


def strehl_ratio(phase_variance: float) -> float:
    """
    Calculates the Strehl ratio.

    Parameters:
        phase_variance: phase variance

    Returns:
        Strehl ratio
    """

    return np.exp(-phase_variance)


def optical_path_difference(opl: np.array, avg_ax: int = 0) -> float:
    """
    Calculates the optical path difference.

    Parameters:
        opl: has to be a numpy array of shape [time, x_axis, y_axis, z_axis]
        avg_ax: axis where average is performed, 0 (default)

    Returns:
        numpy array of the same shape as the optical path length
    """
    if not isinstance(opl, np.ndarray):
        raise ValueError("opl must be a numpy array")

    if avg_ax not in [0, 1, 2, 3]:
        raise ValueError("avg_ax must be one of [0, 1, 2, 3]")
    return opl - np.mean(opl, axis=avg_ax, keepdims=True)


def tropina_aproximation(vibrational_number, rotational_number, molecule):
    electron_mass = s_consts.m_e
    electron_charge = s_consts.e
    spectroscopy_const = constants_tables.spectroscopy_constants(molecule)
    # resonance_distance = omega_gi - omega
    # TODO: Missing implementation
    print("TODO: Missing this implementation")


def buldakov_expansion(
    vibrational_number: int, rotational_number: int, molecule: str
) -> float:
    """
    Calculates the Buldakov expansion

    Parameters:
        vibrational_number: vibrational quantum number (has to be positive)
        rotational_number: rotational quantum number (has to be positive)
        molecule: H2, N2, O2

    Returns:
        buldakov expansion in [m^3]

    Reference:
        Temperature Dependence of Polarizability of Diatomic Homonuclear
        Molecules (https://doi.org/10.1134/BF03355985)
    """
    # Load constants
    spectroscopy_const = constants_tables.spectroscopy_constants(molecule)
    derivative_const = constants_tables.buldakov_polarizability_derivatives_2016(
        molecule
    )
    be_we = spectroscopy_const["B_e"] / spectroscopy_const["omega_e"]

    # Dunham potential energy constants
    (a_0, a_1, a_2) = quantum.potential_dunham_coef_012(molecule)
    a_3 = quantum.potential_dunham_coeff_m(a_1, a_2, 3)

    rotational_degeneracy = rotational_number * (rotational_number + 1)
    vibrational_degeneracy = 2 * vibrational_number + 1

    # Split in terms
    tmp_1 = be_we
    tmp_1 *= -3 * a_1 * derivative_const["first"] + derivative_const["second"]
    tmp_1 *= vibrational_degeneracy
    tmp_1 *= 1 / 2

    tmp_2 = be_we**2
    tmp_2 *= derivative_const["first"]
    tmp_2 *= rotational_degeneracy
    tmp_2 *= 4

    tmp_31a = 7
    tmp_31a += 15 * vibrational_degeneracy**2
    tmp_31a *= a_1**3
    tmp_31a *= -3 / 8

    tmp_31b = 23
    tmp_31b += 39 * vibrational_degeneracy**2
    tmp_31b *= a_2
    tmp_31b *= a_1
    tmp_31b *= 1 / 4

    tmp_31c = 5
    tmp_31c += vibrational_degeneracy**2
    tmp_31c *= a_3
    tmp_31c *= -15 / 4

    tmp_31 = derivative_const["first"] * (tmp_31a + tmp_31b + tmp_31c)

    tmp_32a = 7
    tmp_32a += 15 * vibrational_degeneracy**2
    tmp_32a *= a_1**2
    tmp_32a *= 1 / 8

    tmp_32b = 5
    tmp_32b += vibrational_degeneracy**2
    tmp_32b *= a_2
    tmp_32b * --3 / 4

    tmp_32 = derivative_const["second"] * (tmp_32a + tmp_32b)

    tmp_33 = 7
    tmp_33 += 15 * vibrational_degeneracy**2
    tmp_33 *= a_1
    tmp_33 *= derivative_const["third"]
    tmp_33 *= -1 / 24

    tmp_3 = (tmp_31 + tmp_32 + tmp_33) * be_we**2

    tmp_41 = 1 - a_2
    tmp_41 *= 24
    tmp_41 += 27 * a_1 * (1 + a_1)
    tmp_41 *= derivative_const["first"]

    tmp_42 = 1 + 3 * a_1
    tmp_42 *= derivative_const["second"]
    tmp_42 *= -3

    tmp_43 = 1 / 8 * derivative_const["third"]

    tmp_4 = tmp_41 + tmp_42 + tmp_43
    tmp_4 *= rotational_degeneracy
    tmp_4 *= vibrational_degeneracy
    tmp_4 *= be_we**3

    return derivative_const["zeroth"] + tmp_1 + tmp_2 + tmp_3 + tmp_4


def kerl_polarizability_temperature(
    temperature_K: float, molecule: str, wavelength_nm: float
) -> float:
    """
    Calculates the polarizability using Kerl's extrapolation.

    Parameters:
        temperature_K: reference temperature in [K]
        molecule: H2, N2, O2, Air
        wavelength_nm: signal's wavelength in [nm]

    Returns:
        polarizability in [m^3]

    Reference:
        Polarizability a(w,T,rho) of Small Molecules in the Gas Phase
        (https://doi.org/10.1002/bbpc.19920960517)

    Examples:
        >> kerl_polarizability_temperature(600.0, 'N2', 533.0)
    """
    # Checking cases
    if type(temperature_K) is float and temperature_K < 0:
        raise ValueError("Temperature must be greater than 0 Kelvin!")
    if type(temperature_K) is np.ndarray and (temperature_K < 0).any():
        raise ValueError("Temperature must be greater than 0 Kelvin!")
    if wavelength_nm <= 0:
        raise ValueError("Wavelength must be greater than 0 nanometers!")
    if molecule not in ["Air", "H2", "N2", "O2"]:
        raise ValueError("This function only supports Air, H2, N2 or O2")
    # Check sizes
    mean_const = constants_tables.kerl_interpolation(molecule)
    angular_frequency = 2 * np.pi * s_consts.speed_of_light / (wavelength_nm * 1e-9)

    tmp = mean_const["c"] * temperature_K**2
    tmp += mean_const["b"] * temperature_K
    tmp += 1
    tmp *= mean_const["groundPolarizability"]
    tmp /= 1 - (angular_frequency / mean_const["groundFrequency"]) ** 2

    return tmp  # [m^3]


def atmospheric_index_of_refraction(
    altitude_m: float, vapor_pressure: float = 0.0
) -> float:
    """
    Calculates the atmospheric index of refraction as a function of altitude

    Parameters:
        altitude_m: altitude in [m]
        vapor_pressure: vapor pressure at given altitude in [mbar], 0.0 (default)
        temperature_K: reference temperature in [K]

    Returns:
        index of refraction in [ ]

    Reference:
        The constants in the equation for atmospheric refractive index at radio frequencies (https://ieeexplore.ieee.org/document/4051437)
    """
    atmospheric_prop = Atmosphere(altitude_m)
    temperature = atmospheric_prop.temperature  # [K]
    pressure = atmospheric_prop.pressure * 0.01  # [mbar]
    [K_1, K_2] = constants_tables.smith_atmospheric_constants()

    refractivity = K_2 * vapor_pressure / temperature
    refractivity += pressure
    refractivity *= K_1 / temperature
    refractivity *= 10**-6

    return refractivity + 1


def brewster_angle(
    medium_index_of_refraction: float, vacuum_index_of_refraction: float = 1.0
) -> float:
    """
    Calculates the Brewster angle. Note,
    that medium_index_of_refraction should be greater than the
    vacuum_index_of_refraction

    Parameters:
        medium_index_of_refraction: medium's index of refraction
        vacuum_index_of_refraction: vacuum's index of refraction, 1.0 (default)

    Returns:
        Brewster angle in [degs]
    """
    return np.rad2deg(
        np.arctan(medium_index_of_refraction / vacuum_index_of_refraction)
    )


def total_internal_reflection_angle(
    medium_index_of_refraction: float, vacuum_index_of_refraction: float = 1.0
) -> float:
    """
    Calculates the critical angle that causes total internal reflection. Note,
    that medium_index_of_refraction should be greater than the
    vacuum_index_of_refraction

    Parameters:
        medium_index_of_refraction: medium's index of refraction
        vacuum_index_of_refraction: vacuum's index of refraction, 1.0 (default)

    Returns:
        Critical angle in [degs]
    """
    return np.rad2deg(
        np.arcsin(vacuum_index_of_refraction / medium_index_of_refraction)
    )


def normal_incidence_reflectance(
    medium_index_of_refraction: float, vacuum_index_of_refraction: float = 1.0
) -> float:
    """
    Calculates the reflectance at a normal incidence. Note,
    that medium_index_of_refraction should be greater than the
    vacuum_index_of_refraction

    Parameters:
        medium_index_of_refraction: medium's index of refraction
        vacuum_index_of_refraction: vacuum's index of refraction, 1.0 (default)

    Returns:
       Reflectance at normal incidence in [ ]
    """
    return (
        (medium_index_of_refraction - vacuum_index_of_refraction)
        / (medium_index_of_refraction + vacuum_index_of_refraction)
    ) ** 2


def gladstone_dale_constant(
    mass_density_dict: dict[str, float] = None
) -> dict[str, float]:
    """
    Calculates Gladstone-Dale constants, returns constants haot.constants if
    mass_density_dict is not provided

    Parameters:
        mass density dictionary in [kg/m^3], None (default)

    Returns:
        dict: A dictionary containing
            - Species Gladstone-Dale constants in [m3/kg]
    """
    pol_consts = constants_tables.polarizability()  # [cm^3]

    # Convert polarizability CGS to SI
    pol_SI = {
        key: conversions.polarizability_cgs_to_si(pol_consts[key])
        for key in pol_consts.keys()
    }

    # Calculates species GD
    const_GD = {}
    for key, val in pol_SI.items():
        const_GD[key] = val * s_consts.N_A / molmass.Formula(key).mass
        const_GD[key] /= 2 * s_consts.epsilon_0
        const_GD[key] *= 1e3  # converts [1/g] to [1/kg]

    # Calculate total GD
    if not mass_density_dict:
        return const_GD
    else:
        species_GD = {}
        tot_density = sum(mass_density_dict.values())
        for key in mass_density_dict.keys():
            species_GD[key] = const_GD[key] * mass_density_dict[key] / tot_density

        species_GD["gladstone_dale"] = sum(species_GD.values())
        return species_GD  # [m3/kg]


def gladstone_dale_air_wavelength(wavelength_nm: float) -> float:
    """
    Calculates the Gladstone Dale constant of air using approximation (see reference).

    Parameters:
        wavelength_nm: laser's wavelength [nm]

    Returns:
        Gladstone Dale constant in [m3/kg]

    Reference:
        An Aero-Optical Effect Analysis Method in Hypersonic Turbulence Based on Photon Monte Carlo Simulation (https://doi.org/10.3390/photonics10020172)
    """
    # Equatin (3) from paper
    gd_const = (6.7132 * 1e-8 / wavelength_nm) ** 2
    gd_const += 1

    return 2.2244 * 1e-4 * gd_const  # [m3/kg]



"""
    Date:   10/26/2024
    Author: Martin E. Liza
    File:   constants.py
    Def:    This file contains constants used across the package
"""


def smith_atmospheric_constants() -> tuple[float, float]:
    """
    Constants used to calculate the atmospheric index of refraction

    Returns:
        K_1 in [K]
        K_2 in [K/mbar]

    Reference:
        The constants in the equation for atmospheric refractive index at radio frequencies (https://ieeexplore.ieee.org/document/4051437)
    """
    return [79.0, 4800.0]


def sutherland_constants(gas: str) -> dict[str, float]:
    """
    Constants used for Sutherland's law of viscosity and thermal conductivity


    Parameters:
        gas: Air (default), Argon, N2, O2

    Returns:
        dict: A dictionary containing
            - temperature_ref: reference temperature in [K]
            - viscosity_ref: reference viscosity in [Ns/m2]
            - sutherland_visc: Sutherland viscosity constant in [K]
            - conductivity_ref: reference thermal conductivity in [W/mK]
            - sutherland_cond: Sutherland thermal conductivity in [K]

    Reference:
        Viscous Fluid Flow, International Edition, 4th (White F., ISBN 978 1 260
        59786). Table 1.1 and Table 1.2
    """

    dict_out = {}
    if gas == "Air":
        dict_out["temperature_ref"] = 273.0  # [K]
        dict_out["viscosity_ref"] = 1.716e-5  # [kg/ms]
        dict_out["sutherland_visc"] = 111.0  # [K]
        dict_out["conductivity_ref"] = 0.0241  # [W/mK]
        dict_out["sutherland_cond"] = 194.0  # [K]

    if gas == "Argon":
        dict_out["temperature_ref"] = 273.0  # [K]
        dict_out["viscosity_ref"] = 2.125e-5  # [kg/ms]
        dict_out["sutherland_visc"] = 114.0  # [K]
        dict_out["conductivity_ref"] = 0.0163  # [W/mK]
        dict_out["sutherland_cond"] = 170.0  # [K]

    if gas == "N2":
        dict_out["temperature_ref"] = 273.0  # [K]
        dict_out["viscosity_ref"] = 1.663e-5  # [kg/ms]
        dict_out["sutherland_visc"] = 107.0  # [K]
        dict_out["conductivity_ref"] = 0.0242  # [W/mK]
        dict_out["sutherland_cond"] = 150.0  # [K]

    if gas == "O2":
        dict_out["temperature_ref"] = 273.0  # [K]
        dict_out["viscosity_ref"] = 1.919e-5  # [kg/ms]
        dict_out["sutherland_visc"] = 139.0  # [K]
        dict_out["conductivity_ref"] = 0.0244  # [W/mK]
        dict_out["sutherland_cond"] = 240.0  # [K]

    return dict_out


def karl_2003() -> dict[str, float]:
    """
    Gladstone-Dale constants used in Karl's experiments

    Returns:
        dict: A dictionary containing
            - species in units of [m^3/kg]

    Reference:
        High Enthalpy Cylinder Flow in HEG, A Basis for CFD Validation (https://arc.aiaa.org/doi/pdf/10.2514/6.2003-4252)
    """

    dict_out = {
        "N": 3.01e-4,
        "O": 1.82e-4,
        "NO": 2.21e-4,
        "N2": 2.38e-4,
        "O2": 1.90e-4,
    }  # [m^3/kg]
    return dict_out


def polarizability() -> dict[str, float]:
    """
    Polarizabilities at laboratory conditions

    Returns:
        dict: A dictionary of polarizabilities
            - Volumetric species polarizability in CGS units of [cm^3]

    Reference:
        Handbook of Chemistry and Physics, 95th edition
        (https://doi.org/10.1201/b17118)

        Optical spectroscopy of high L n=10 Rydberg states of nitrogen
        (https://doi.org/10.1103/PhysRevA.54.314)

        A numerical study of coupled Hartree-Fock theory for open-shell systems
        (https://doi.org/10.1080/00268977500102811)

        Ab initio calculations of the properties of NO+ in its ground
        electronic state X 1Sigma+
        (https://doi.org/10.1016/0009-2614(93)89356-M)

        Analysis of the 8f, 9f, and 10f, v=1 Rydberg states of N2
        (https://doi.org/10.1103/PhysRevA.44.3007)
    """
    dict_out = {
        "N+": 0.559e-24,
        "O+": 0.345e-24,
        "NO+": 1.021e-24,
        "N2+": 2.386e-24,
        "O2+": 0.238e-24,
        "N": 1.100e-24,
        "O": 0.802e-24,
        "NO": 1.700e-24,
        "N2": 1.7403e-24,
        "O2": 1.5689e-24,
    }  # [cm^3]
    return dict_out


def buldakov_polarizability_derivatives_2016(molecule: str) -> dict[str, float]:
    """
    Polarizability derivative constants

    Parameters:
        molecule: H2, N2, O2

    Reference:
        Temperature dependence of polarizability of diatomic homonuclear
        molecules (https://doi.org/10.1134/BF03355985)

        Handbook of Chemistry and Physics, 95th edition
        (https://doi.org/10.1201/b17118)

        An accurate calculation of the polarizability of the hydrogen molecule
        and its dependence on rotation, vibration and isotopic substitution
        (https://doi.org/10.1080/00268978000103191)

        Theoretical study of the effects of vibrational‐rotational interactions
        on the Raman spectrum of N2 (https://doi.org/10.1063/1.445482)

        Frequency-dependent polarizability of O2 and van der Waals coefficients
        of dimers containing O2 (https://doi.org/10.1063/1.467256)
    """

    dict_out = {}
    if molecule == "H2":
        dict_out["zeroth"] = 0.7849e-30
        dict_out["first"] = 0.90e-30
        dict_out["second"] = 0.49e-30
        dict_out["third"] = -0.85e-30

    if molecule == "N2":
        dict_out["zeroth"] = 1.7801e-30
        dict_out["first"] = 1.86e-30
        dict_out["second"] = 1.2e-30
        dict_out["third"] = -4.6e-30

    if molecule == "O2":
        dict_out["zeroth"] = 1.6180e-30
        dict_out["first"] = 1.76e-30
        dict_out["second"] = 3.4e-30
        dict_out["third"] = -23.7e-30

    return dict_out  # [m^3]


def kerl_interpolation(molecule: str) -> dict[str, float]:
    """
    Constants used in Kerl's extrapolation method

    Parameters:
        molecules: H2, N2, O2, Air

    Returns:
        dict: A dictionary containing
            - groundPolarizability: polarizability at ground level in [m^3]
            - groundFrequency: frequency at ground level in [Hz]
            - b: extrapolation constant in [1/K]
            - c: extrapolation constant in [1/K^2]

    Reference:
        Polarizability a(w,T,rho) of Small Molecules in the Gas Phase
        (https://doi.org/10.1002/bbpc.19920960517)
    """

    dict_out = {}
    if molecule == "H2":
        dict_out["groundPolarizability"] = 0.80320e-30  # [m^3]
        dict_out["groundFrequency"] = 2.1399e16  # [1/s]
        dict_out["b"] = 5.87e-6  # [1/K]
        dict_out["c"] = 7.544e-9  # [1/K^2]
    if molecule == "N2":
        dict_out["groundPolarizability"] = 1.7406e-30  # [m^3]
        dict_out["groundFrequency"] = 2.6049e16  # [1/s]
        dict_out["b"] = 1.8e-6  # [1/K]
        dict_out["c"] = 0.0  # [1/K^2]
    if molecule == "O2":
        dict_out["groundPolarizability"] = 1.5658e-30  # [m^3]
        dict_out["groundFrequency"] = 2.1801e16  # [1/s]
        dict_out["b"] = -2.369e-6  # [1/K]
        dict_out["c"] = 8.687e-9  # [1/K^2]
    if molecule == "Air":
        dict_out["groundPolarizability"] = 1.6970e-30  # [m^3]
        dict_out["groundFrequency"] = 2.47044e16  # [1/s]
        dict_out["b"] = 10.6e-6  # [1/K]
        dict_out["c"] = 7.909e-9  # [1/K^2]

    return dict_out


def spectroscopy_constants(molecule: str) -> dict[str, float]:
    """
    Returns spectroscopy constants

    Parameters:
        molecule: NO+, N2+, O2+, NO, N2, O2, H2

    Returns:
        dict: A dictionary containing spectroscopy constants
            - omega_e: vibrational constant - first term in [cm^-1]
            - omega_xe: vibrational constant – second term in [cm^-1]
            - omega_ye: vibrational constant – third term in [cm^-1]
            - B_e: equilibrium rotational constant in [cm^-1]
            - alpha_e: rotational constant in [cm^-1]
            - D_e: centrifugal distortion in [cm^-1]
            - r_e: internuclear distance in [m]

    Reference:
        NO+ https://webbook.nist.gov/cgi/cbook.cgi?Name=NO%2B&Units=SI&cDI=on

        N2+ https://webbook.nist.gov/cgi/cbook.cgi?Name=N2%2B&Units=SI&cDI=on

        O2+ https://webbook.nist.gov/cgi/cbook.cgi?Name=O2%2B&Units=SI&cDI=on

        NO https://webbook.nist.gov/cgi/cbook.cgi?Name=NO&Units=SI&cDI=on

        N2 https://webbook.nist.gov/cgi/cbook.cgi?Name=N2&Units=SI&cDI=on

        O2 https://webbook.nist.gov/cgi/cbook.cgi?Name=O2&Units=SI&cDI=on

        H2 https://webbook.nist.gov/cgi/cbook.cgi?Name=H2&Units=SI&cDI=on
    """
    dict_out = {}

    if molecule == "NO+":
        dict_out["omega_e"] = 2376.72
        dict_out["omega_xe"] = 16.255
        dict_out["omega_ye"] = -0.01562
        dict_out["B_e"] = 1.997195
        dict_out["alpha_e"] = 0.018790
        dict_out["D_e"] = 6.64e-6
        dict_out["r_e"] = 1.06322e-10

    if molecule == "N2+":
        dict_out["omega_e"] = 2207.0115
        dict_out["omega_xe"] = 16.0616
        dict_out["omega_ye"] = -0.04289
        dict_out["B_e"] = 1.93176
        dict_out["alpha_e"] = 0.0181
        dict_out["D_e"] = 6.10e-6
        dict_out["r_e"] = 1.11642e-10

    if molecule == "O2+":
        dict_out["omega_e"] = 1905.892
        dict_out["omega_xe"] = 16.489
        dict_out["omega_ye"] = 0.02057
        dict_out["B_e"] = 1.689824
        dict_out["alpha_e"] = 0.019363
        dict_out["D_e"] = 5.32e-6
        dict_out["r_e"] = 1.1164e-10

    if molecule == "NO":
        dict_out["omega_e"] = 1904.1346
        dict_out["omega_xe"] = 14.08836
        dict_out["omega_ye"] = 0.01005
        dict_out["B_e"] = 1.704885
        dict_out["alpha_e"] = 0.0175416
        dict_out["D_e"] = 0.54e-6
        dict_out["r_e"] = 1.15077e-10

    if molecule == "N2":
        dict_out["omega_e"] = 2358.57
        dict_out["omega_xe"] = 14.324
        dict_out["omega_ye"] = -2.26e-3
        dict_out["B_e"] = 1.998241
        dict_out["alpha_e"] = 0.017318
        dict_out["D_e"] = 5.76e-6
        dict_out["r_e"] = 1.2126e-10

    if molecule == "O2":
        dict_out["omega_e"] = 1580.161
        dict_out["omega_xe"] = 11.95127
        dict_out["omega_ye"] = 0.0458489
        dict_out["B_e"] = 1.44562
        dict_out["alpha_e"] = 0.0159305
        dict_out["D_e"] = 4.839e-6
        dict_out["r_e"] = 1.20752e-10

    if molecule == "H2":
        dict_out["omega_e"] = 4401.21
        dict_out["omega_xe"] = 121.33
        dict_out["omega_ye"] = 0.0
        dict_out["B_e"] = 60.853
        dict_out["alpha_e"] = 3.062
        dict_out["D_e"] = 0.0471
        dict_out["r_e"] = 0.74144

    return dict_out
